--!parent
--!encoding=utf-8

--****************************************************************************
-- @file			LSW CORE LIB: URL
-- @author		Valentin Schmidt
-- @version	 0.2
-- @desc			URL/HTTP related utilitiy functions
--****************************************************************************

property $

property _base64LookupEnc
property _base64LookupDec

----------------------------------------
-- @constructor
----------------------------------------
on new (me)

	-- libs
	$.import("string")
	
	me._base64LookupEnc = [65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,\
	85,86,87,88,89,90,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,\
	114,115,116,117,118,119,120,121,122,48,49,50,51,52,53,54,55,56,57,43,47]

	me._base64LookupDec = []
	repeat with k = 1 to 256
		me._base64LookupDec[k] = me._base64LookupEnc.findpos(k)-1
	end repeat

	return me
end

----------------------------------------
-- URL encodes a string (as specified in RFC 3986)
-- @static
-- @param {string} str
-- @param {bool} [raw=false] - If FALSE or VOID, SPACE is encoded as "+"
-- @return {string}
----------------------------------------
on urlencode (me, str, raw, keepSlashes)
	allowed = "abcdefghijklmnopqrstuvwxyz0123456789-_."
	res = ""
	len = str.length
	repeat with i = 1 to len
		c = str.char[i]
		
		if (c=" " AND raw<>TRUE) then 
			c="+"
		else if (c="/" AND keepSlashes=TRUE) then 
			-- noting
		else if not (allowed contains c) then
			c = me._hexencode(c)
		end if
		
		put c after res
	end repeat
	return res
end

----------------------------------------
-- URL decodes a string
-- @static
-- @param {string} str
-- @return {string}
----------------------------------------
on urldecode (me, str)
	res = ""
	ba = bytearray()
	len = str.length
	repeat with i = 1 to len
		c = str.char[i]
		if (c = "%") then
			ba.writeInt8(rgb(str.char[i+1..i+2]).blue)
			i = i + 2
		else if (c = "+") then
			ba.writeInt8(32)
		else
			ba.writeInt8(chartonum(c))
		end if
	end repeat
	ba.position = 1
	return ba.readRawString(ba.length)
end

----------------------------------------
-- Parses string like a query string, returns propList
-- @static
-- @param {string} str
-- @return {propList}
----------------------------------------
on parse_str (me, str)
	res = [:]
	parts = $.string.explodeBySingleChar("&", str)
	repeat with p in parts
		tmp = $.string.explodeBySingleChar("=", p)
		if count(tmp)>1 then 
			--res[tmp[1]] = me.urldecode(tmp[2])
			res.addProp(tmp[1], me.urldecode(tmp[2]))
		else res[tmp[1]] = ""
	end repeat
	return res
end

----------------------------------------
-- Parses URL
-- 'http://username:password@hostname:88/path?arg=value#anchor'
-- @static
-- @param {string} str
-- @return {propList}
----------------------------------------
on parse_url (me, str)
	res = [:]
	
	-- scheme
	parts = $.string.explode("://", str, 2)
	if parts.count>1 then 
		res[#scheme] = parts[1]
		str = parts[2]
	else
		res[#scheme] = ""
	end if
	
	-- username and password
	parts = $.string.explodeBySingleChar("@", str, 2)
	if parts.count>1 then
		str = parts[2]
		res[#user] = parts[1]
		parts = $.string.explodeBySingleChar(":",	parts[1], 2)
		if parts.count>1 then
			res[#user] = parts[1]
			res[#pass] = parts[2]
		else
			res[#pass] = ""
		end if
	else
		res[#user] = ""
		res[#pass] = ""
	end if
	
	-- fragment (anchor)
	parts = $.string.explodeBySingleChar("#", str, 2)
	if parts.count>1 then
		res[#fragment] = parts[2]
		str = parts[1]
	else
		res[#fragment] = ""
	end if
	
	-- query
	parts = $.string.explodeBySingleChar("?", str, 2)
	if parts.count>1 then
		res[#query] = parts[2]
		str = parts[1]
	else
		res[#query] = ""
	end if
	
	-- path
	res[#file] = ""
	parts = $.string.explodeBySingleChar("/", str, 2)
	if parts.count>1 then
		res[#path] = "/"&parts[2]
		if the last char of str<>"/" then
			tmp = $.string.explodeBySingleChar("/", str)
			res[#file] = tmp[tmp.count]			
		end if
		str = parts[1]
	else
		res[#path] = "/"
	end if
	
	-- host and port
	parts = $.string.explodeBySingleChar(":", str, 2)
	if parts.count>1 then
		res[#host] = parts[1]
		res[#port] = parts[2]
	else
		res[#host] = str
		res[#port] = ""
	end if
	
	return res
end

----------------------------------------
-- Generates URL-encoded query string
-- @static
-- @param {propList} param_obj
-- @param {bool} [raw=false]
-- @return {string}
----------------------------------------
on http_build_query (me, param_obj, raw)
	query_str = ""
	cnt = count(param_obj)
	repeat with i = 1 to cnt
		put param_obj.getPropAt(i)&"="&me.urlencode(param_obj[i], raw)&"&" after query_str
	end repeat
	delete the last char of query_str
	return query_str
end

----------------------------------------
-- ByteArray -> ASCII String
-- @param {string|bytearray} baData
-- @return {string}
----------------------------------------
on base64_encode (me, baData)
	if stringP(baData) then baData = byteArray(baData)
	baOut = byteArray()
	i=1
	strRemainder = baData.length mod 3
	strLength		= baData.length - strRemainder
	repeat while (i < strLength)
		a = baData[i]
		b = baData[i+1]
		c = baData[i+2]
		byte1 = me._base64LookupEnc[ ( bitAnd( a, 252 ) / 4 ) + 1]
		byte2 = me._base64LookupEnc[ ( bitAnd( (a*256)+b, 1008 ) / 16 ) + 1]
		byte3 = me._base64LookupEnc[ ( bitAnd( (b*256)+c, 4032 ) / 64 ) + 1]
		byte4 = me._base64LookupEnc[ ( bitAnd( c, 63 ) ) + 1]
		baOut.writeInt32( byte1 + byte2*256 + byte3*65536 + byte4*16777216)
		i = i + 3
	end repeat
	if (strRemainder = 1) then
		a = baData[i]
		baOut.writeInt8( me._base64LookupEnc[ ( bitAnd( a, 252 ) / 4 ) + 1] )
		baOut.writeInt8( me._base64LookupEnc[ ( bitAnd( a*256, 1008 ) / 16 ) + 1] )
		baOut.writeRawString( "==", 2)
	else if (strRemainder = 2) then
		a = baData[i]
		b = baData[i+1]
		baOut.writeInt8( me._base64LookupEnc[ ( bitAnd( a, 252 ) / 4 ) + 1] )
		baOut.writeInt8( me._base64LookupEnc[ ( bitAnd( (a*256)+b, 1008 ) / 16 ) + 1] )
		baOut.writeInt8( me._base64LookupEnc[ ( bitAnd( b*256, 4032 ) / 64 ) + 1] )
		baOut.writeRawString( "=", 1)
	end if
	baOut.position = 1
	return baOut.readRawString(baOut.length)
end

----------------------------------------
-- ASCII String -> ByteArray
-- @param {string} str
-- @return {bytearray}
----------------------------------------
on base64_decode (me, str)
	baIn = byteArray(str)
	baOut = byteArray()
	len = baIn.length
	i		= 1
	iEnd = integer(((len)/4.0)-0.5)*4
	repeat while (i < iEnd)
		a = me._base64LookupDec[baIn[i]]
		b = me._base64LookupDec[baIn[i+1]]
		c = me._base64LookupDec[baIn[i+2]]
		d = me._base64LookupDec[baIn[i+3]]
		byte1 = bitAnd( (a*64)+b, 4080 ) / 16
		byte2 = bitAnd( (b*64)+c, 1020 ) / 4
		baOut.writeInt16( byte1 + byte2*256)
		baOut.writeInt8( bitAnd( (c*64)+d, 255 ) )
		i = i + 4
	end repeat
	if (len-iEnd = 3) then
		a = me._base64LookupDec[baIn[i+1]]
		b = me._base64LookupDec[baIn[i+2]]
		c = me._base64LookupDec[baIn[i+3]]
		baOut.writeInt8( bitAnd( (a*64)+b, 4080 ) / 16 )
		baOut.writeInt8( bitAnd( (b*64)+c, 1020 ) / 4	)
	else if (len-iEnd = 2) then
		a = me._base64LookupDec[baIn[i+1]]
		b = me._base64LookupDec[baIn[i+2]]
		baOut.writeInt8( bitAnd( (a*64)+b, 4080 ) / 16 )
	end if
	return baOut
end

--**************************************
-- PRIVATE
--**************************************

----------------------------------------
-- Returns sequence of hex codes (prepended with "%") for specified UTF-8 string
-- @param {string|bytearray} ba
-- @return {string}
----------------------------------------
on _hexencode (me, ba)
	res = ""
	if stringP(ba) then ba = bytearray(ba)
	str = ba.toHexString(1,ba.length)
	cnt = str.word.count
	repeat with i = 1 to cnt
		put "%" & str.word[i] after res
	end repeat
	return res
end
